class Graph {
  origin;
  
  constructor() {
    this.origin = null;
    this.axes = [];//a pair of vectors that generate the unit cell of the periodic graph.
    this.lattice = [];//generated by the axes and origin
    
    this.vertices = [];//cartesian coordinates for every vertex drawn
    this.vorbits = [];//the orbit or family that each vertex belongs to (represented by the same color)
    this.vshifts = [];//indicates which cell each vertex copy is in. Vertices inside the cell that is place at the origin have shift (0,0). Vertices in the cell to the right of it have shift (1,0)
    
    this.faces = [];//each face is an ordered colllection of vertices 
    this.forbits = [];
    
    this.fVReps = [];
    this.fShiftsReps = [];
    
    this.vertexRepsFrac = []; //vertex representatives in fractional coordinates
    
    this.edges = [];//all edges drawn. Each edge is a pair of vertices (u,v) where u,v are in this.vertices
    this.eorbits = [];//the orbit or family that each edge belongs to (represented by the same color)
    
    this.edgeReps = []; //pairs of vertex orbits; not pairs of vertex IDs 
    this.eRepShifts = []; //for an edge (u,v) the edge shift is vshift - ushift. The direction of the edge affects the shift
    
    //To draw the span graph (version of the graph where only one copy of each edge orbit is drawn)
    //In the span graph, only the vertices of a single cell are drawn along with vertices necessary for the edges
    this.spanV = [];
    this.spanVOrbits = [];
    this.spanVShifts = [];
    this.spanE = [];
    
    //Index of last clicked vertex
    this.selectedVID = -1
    //vertices selected for polygon drawing
    this.allSelected = [];
  }
  reset() {
    this.resetUCInfo();
    this.resetVInfo();
    this.resetEInfo();
    this.resetSpanInfo();
    this.resetFInfo();
    this.resetSelected();
  }
  resetSelected() {
    this.selectedVID = -1;
    this.allSelected = [];
  }
  resetFInfo() {
    this.faces = [];
    this.forbits = [];
  }
  resetUCInfo() {
    this.origin = null;
    this.axes = [];
    this.lattice = [];
  }
  resetVInfo() {
    this.vertices = [];
    this.vorbits = [];
    this.vshifts = [];
    this.vertexRepsFrac = [];
  }
  resetEInfo() {
    this.edges = [];
    //this.elengths = [];
    this.eorbits = [];
    this.eRepShifts = [];
    this.edgeReps = [];
  }
  resetSpanInfo() {
    this.spanV = [];
    this.spanVOrbits = [];
    this.spanVShifts = [];
    this.spanE = [];
  }
  /*========================================================================*/
  /*============================ SPANNING GRAPH ============================*/
  /*========================================================================*/
  //In a span graph, you draw the minimal number of edges needed to represent a periodic graph
  computeSpan() {
    this.spanV = [];
    this.spanVOrbits = [];
    this.spanVShifts = [];
    this.spanE = [];
    
    this.spanV = fracToCartCoords(this, this.vertexRepsFrac);
    
    for (let i = 0; i < this.vertexRepsFrac.length; i++) {
      this.spanVOrbits.push(i);
      this.spanVShifts.push(new Coordinate(0,0));
    }
    
    for (let i = 0; i < this.edgeReps.length; i++) {
      let shift = this.eRepShifts[i];
      let edge = this.edgeReps[i];
      if (isZeroShift(shift)) {
        this.spanE.push(edge);
      }
      else {
        var newFracPt = new Coordinate(this.vertexRepsFrac[edge[1]].x, this.vertexRepsFrac[edge[1]].y);
        newFracPt.add(shift);
        this.spanV.push(fracToCartCoord(this, newFracPt));
        this.spanVOrbits.push(edge[1]);
        this.spanVShifts.push(shift);
        this.spanE.push([edge[0], this.spanV.length - 1]);
      }
    }
  }
  /*========================================================================*/
  /*================================= AXES =================================*/
  /*========================================================================*/
  createAxis(coordinate) {
    var axis = new Coordinate(coordinate.x, coordinate.y);
    axis.subtract(this.origin);
    if (this.axes.length < 2) {
      this.axes.push(axis);
      stateHandler();
    }
  }
  moveAxis(axisIndex, coordinate, fSize) {
    this.axes[axisIndex].copy(coordinate);
    this.axes[axisIndex].subtract(this.origin);
    if (this.vertexRepsFrac.length != 0) {
      this.makeNewFragment(fSize, false);
      this.computeSpan();
    }
    if (this.axes.length == 2) {
      this.computeLattice(fSize);
      stateHandler();
      printGraph(graph);
    }
  }
  /*======================================================================================*/
  /*================================= ORIGIN AND LATTICE =================================*/
  /*======================================================================================*/
  createOrigin(coordinate) {
    this.origin = new Coordinate(coordinate.x,coordinate.y);
    stateHandler();
    //alert(typeof(this.origin));
  }
  //Move origin or translate graph
  moveOrigin(coordinate, fSize) {
    var translation = new Coordinate(coordinate.x, coordinate.y);
    translation.subtract(this.origin);
    this.origin.copy(coordinate);
    if (this.vertices.length !== 0) {
      translatePoints(this.vertices, translation);
      translatePoints(this.lattice, translation);
      this.computeSpan();
    }
    if (this.axes.length == 2)
      this.computeLattice(fSize);
    //console.log("origin: " + this.origin);
  }
  computeLattice(fSize) {
    this.lattice = [];
    this.computeVertexCopies(this.origin, -1, fSize);
  }
  /*============================================================================*/
  /*================================= VERTICES =================================*/
  /*============================================================================*/
  createVertexOrbit(coordinate, fSize) {
    var orbit = 0;
    var size_vorbits = this.vorbits.length;
    if (size_vorbits != 0) {
      orbit = this.vorbits[size_vorbits - 1] + 1;
    }
    this.vertexRepsFrac.push(cartToFracCoord(this, coordinate, true));
    //if there are no edges in the graph, we add a copy of the new vertex to the end of spanV
    if (this.vertexRepsFrac.length === this.spanV.length + 1) {
      this.spanV.push(fracToCartCoord(this, this.vertexRepsFrac[this.vertexRepsFrac.length - 1]));
      this.spanVOrbits.push(this.vertexRepsFrac.length - 1);
      this.spanVShifts.push(new Coordinate(0,0));
    }
    //if there are edges in the graph, we want vertices created by edges with non zero shifts to be at the end of spanV
    else {
      this.computeSpan();
    }
    
    if (size_vorbits >= VERTEX_COLORS.length) {
      //https://stackoverflow.com/questions/1152024/best-way-to-generate-a-random-color-in-javascript
      var randomColor = '#' + (0x1000000 + Math.floor(Math.random() * 0x1000000)).toString(16).substr(1);
      VERTEX_COLORS.push(randomColor);
    }
    
    var shift = getVertexShift(this, coordinate)
    this.computeVertexCopies(getVertexRepCoord(this, coordinate), orbit, fSize);
    stateHandler();
  }
  //Assuming coordinate is the representative that lies at shift (0,0)
  computeVertexCopies(coordinate, orbit, fSize) {
    for (let i = Number(fSize.minx); i <= Number(fSize.maxx); i++) {
      const xaxisDisp = new Coordinate(this.axes[0].x, this.axes[0].y);
      xaxisDisp.multiply(Number(i));
      for (let j = Number(fSize.miny); j <= Number(fSize.maxy); j++) {
        const yaxisDisp = new Coordinate(this.axes[1].x, this.axes[1].y);
        yaxisDisp.multiply(Number(j));
        const newcoords = new Coordinate(coordinate.x, coordinate.y);
        newcoords.add(xaxisDisp);
        newcoords.add(yaxisDisp);
        if (orbit !== -1) {
          this.vertices.push(newcoords);
          this.vorbits.push(orbit);
          this.vshifts.push(new Coordinate((i),(j)));
        }
        else
          this.lattice.push(newcoords);
      }
    }
  }
  moveVertexOrbit(vertexIndex, coordinate) {
    var orbit = this.vorbits[vertexIndex];
    var nrV = this.vertices.length;
    var translation;
    
    var xFactor;
    var yFactor;
    //Added a for loop because otherwise the point moves really slow near the corner
    for (let k = 0; k < 100 ; k++) { 
      if (coordinate.x > this.vertices[vertexIndex].x) 
        xFactor = 2;
      else if (Math.round(coordinate.x) === Math.round(this.vertices[vertexIndex].x))
        xFactor = 0;
      else xFactor = -2;
      if (coordinate.y > this.vertices[vertexIndex].y) 
        yFactor = 2;
      else if (Math.round(coordinate.y) === Math.round(this.vertices[vertexIndex].y))
        yFactor = 0;
      else yFactor = -2;
      
      //Case 0: If the new point position is in bounds
      if (shiftIsEqual(this.vshifts[vertexIndex], getVertexShift(this, coordinate))) {
        translation = new Coordinate(coordinate.x, coordinate.y);
        translation.subtract(this.vertices[vertexIndex]);
      }
      //The rest of the cases are for when the new point position is out of bounds
      //Case 1: If the point can be moved in the x and y direction of the new point position, move it a little in x and y.
      else if (shiftIsEqual(this.vshifts[vertexIndex], getVertexShift(this, new Coordinate(this.vertices[vertexIndex].x + xFactor, this.vertices[vertexIndex].y + yFactor)))) {
        translation = new Coordinate(xFactor, yFactor);
      }
      //Case 2: If the point can be moved in the x direction, move it a little in x.
      else if (shiftIsEqual(this.vshifts[vertexIndex], getVertexShift(this, new Coordinate(this.vertices[vertexIndex].x + xFactor, this.vertices[vertexIndex].y)))) {
        translation = new Coordinate(xFactor, 0);
      }
      //Case 3: If the point can be moved in the y direction, move it a little in y.
      else if (shiftIsEqual(this.vshifts[vertexIndex], getVertexShift(this, new Coordinate(this.vertices[vertexIndex].x, this.vertices[vertexIndex].y + yFactor)))) {
        translation = new Coordinate(0, yFactor);
      }
      //Extra cases
      //Case 3: Sometimes you want to move the point in the x direction of the goal even if its y coordinate gets further from the goal.
      //For example consider having a left-leaning rhombus cell. If your point is at the left bound and your goal position is below and
      //to the left of the cell, then you would want the point to move to the bottom left corner.
      else if (shiftIsEqual(this.vshifts[vertexIndex], getVertexShift(this, new Coordinate(this.vertices[vertexIndex].x + xFactor, this.vertices[vertexIndex].y - yFactor)))) {
        translation = new Coordinate(xFactor, -yFactor);
      }
      //Case 4: Sometimes you want to move the point in the y direction of the goal even if its x coordinate gets further from the goal.
      else if (shiftIsEqual(this.vshifts[vertexIndex], getVertexShift(this, new Coordinate(this.vertices[vertexIndex].x - xFactor, this.vertices[vertexIndex].y + yFactor)))) {
        translation = new Coordinate(-xFactor, yFactor);
      }
      else {
        return;
      }
      
      for (let i = 0; i < nrV; i++) {
        if (this.vorbits[i] === orbit) {
          const newcoords = new Coordinate(this.vertices[i].x, this.vertices[i].y);
          newcoords.add(translation);
          this.vertices[i].copy(newcoords);
        }
      }
      this.vertexRepsFrac[this.vorbits[vertexIndex]] = cartToFracCoord(this, this.vertices[vertexIndex], true);
    }
    this.computeSpan();
    printGraph(this);
  }
  deleteVertexOrbit(selected, fSize) {
    var orbit = this.vorbits[selected];
    
    this.removeVOrbitFromEdgeData(orbit, fSize);
    this.removeVOrbitFromEdgeRepData(orbit);
    this.removeVOrbitFromVertexData(orbit);
    this.computeSpan();

    printGraph(this);
  }
  
  /*===============================================================================*/
  /*==================================== EDGES ====================================*/
  /*===============================================================================*/
  //Keep in mind that vertex copies in an orbit are created bottom-up left to right.
  //We can use that index information when creating edges.
  createEdgeOrbit(headVertexId, tailVertexId, fSize) {
    if (headVertexId === tailVertexId) return;
    
    /* Edge Rep Data*/
    var eshift = getEdgeShift(this, headVertexId, tailVertexId);
    var headOrbit = this.vorbits[headVertexId];
    var tailOrbit = this.vorbits[tailVertexId];
    
    /* If the edge orbit already exists, return ; This avoids multigraphs */
    if (this.inEdgeReps([headOrbit,tailOrbit], eshift)) {
      return;
    }
    
    this.eRepShifts.push(eshift);
    this.edgeReps.push([headOrbit,tailOrbit]);
    if (EDGE_COLORS.length < this.edgeReps.length)
          EDGE_COLORS.push('#' + (0x1000000 + Math.floor(Math.random() * 0x1000000)).toString(16).substr(1));
    this.computeSpan();
    
    /* Edge Orbit Data*/
    var eorbit = 0;
    var size_eorbits = this.eorbits.length;
    if (size_eorbits != 0) {
      eorbit = this.eorbits[size_eorbits - 1] + 1;
    }
      
    this.computeEdgeCopies(headOrbit, tailOrbit, eshift, eorbit, fSize);
  }
  computeEdgeCopies(headOrbit, tailOrbit, eshift, eorbit, fSize) {
    var rangeY = fSize.maxy - fSize.miny + 1;
    var rangeX = fSize.maxx - fSize.minx + 1;
    var headStartId = this.vorbits.findIndex((element) => element === headOrbit);
    var tailStartId = this.vorbits.findIndex((element) => element === tailOrbit);
    var headCurrId = headStartId; 
    var tailCurrId = tailStartId + (eshift.x * rangeY) + eshift.y;//the tail vertex copy that is eshift away from headCurrId
    //suppose the edge rep is vertex orbit pair (0,0) with shift (-1, 0).
    //then tailCurrId above would be -rangeY which is a negative number meaning the index is out of bounds.
    //To deal with this, we have the following if statement
    if (tailCurrId < tailStartId) {
      tailCurrId = tailStartId;
      headCurrId = headStartId + (eshift.x * rangeY * (-1)) + (eshift.y * (-1));
    }
    for (let i = 0; i < rangeX * rangeY; i++) {
      if (this.vorbits[headCurrId] !== headOrbit || this.vorbits[tailCurrId] !== tailOrbit) {
        break;
      }
      //Make an edge if the pair of vertices has the shift we want
      if (shiftSumIsEqual(eshift, this.vshifts[headCurrId], this.vshifts[tailCurrId])) {
        this.edges.push([headCurrId, tailCurrId]);
        this.eorbits.push(eorbit);
      }
      headCurrId++;
      tailCurrId++;
    }
  }
  
  deleteEdgeOrbit(previousSelected, selected) {
    var orbit = getEOrbit(this, previousSelected, selected);
    if (orbit === -1) {
      console.log("Could not find the edge representative for deletion");
      return;
    }
    this.removeEOrbitFromEdgeData(orbit);
    this.computeSpan();
    printGraph(this);
  }
  
  removeEOrbitFromEdgeData(orbit) {
    for (let i = this.edges.length - 1; i > -1; i--) {
      if (this.eorbits[i] === orbit) {
        this.eorbits.splice(i, 1);
        this.edges.splice(i, 1);
      }
    }
    this.edgeReps.splice(orbit,1);
    this.eRepShifts.splice(orbit,1);
    this.reassignEdgeOrbits();
    
    var orbitColor = EDGE_COLORS[orbit];
    for (let i = orbit; i <= this.edgeReps.length; i++) {
      EDGE_COLORS[i] = EDGE_COLORS[i+1];
    }
    EDGE_COLORS[this.edgeReps.length] = orbitColor;
  }
  /*============================================================================*/
  /*================================= FACES =================================*/
  /*============================================================================*/
  createFaceOrbit(fvertexIds, fSize) {
    //if face already exists, do not add
    //TODO
    
    //Make representative
    var fVRep = [];
    var fShiftsRep = [];
    for (let i = 0; i < fvertexIds.length; i++) {
      fVRep.push(this.vorbits[fvertexIds[i]]);
      fShiftsRep.push(this.vshifts[fvertexIds[i]]);
    }
    fVReps.push(fVRep);
    fShiftsReps.push(fShiftsRep);
   
    
    //make copies
  }
  equalVOrbits(vertexIds1, vertexIds2) {
    for (let i = 0; i < vertexIds1.length; i++) {
      if (this.vorbits[vertexIds1[i]] != this.vorbits[vertexIds2[i]])
        return false;
    }
    return true;
  }
  getShiftDifference(vertexIds) {
    var startShift = new Coordinate(0,0);
    startShift = new Coordinate(this.vshifts[vertexIds[0]].x, this.vshifts[vertexIds[0]].y);
    var shiftDifferences = [];
    
    for (let i = 0; i < vertexIds.length; i++) {
      var shift = new Coordinate(0,0);
      shift.copy(this.vshifts[vertexIds[i]]);
      shift.subtract(startShift);
      shiftDifferences.push(shift);
    }
    return shiftDifferences;
  }
  equalShiftDifferences(shiftDifferences, newVertexIds) {
    var newShiftDifferences = this.getShiftDifference(newVertexIds);
    for (let i = 0; i < shiftDifferences.length; i++)
      if (!shiftIsEqual(shiftDifferences[i], newShiftDifferences[i]))
        return false;
    return true;
  }
  faceExists(vertexIds) {
    
    
  }
  computeFaceCopies(vertexIds, fSize) {
    var rangeY = fSize.maxy - fSize.miny + 1;
    var rangeX = fSize.maxx - fSize.minx + 1;
    var orbit;
    if (this.forbits.length === 0)
      orbit = 0;
    else
      orbit = this.forbits[this.forbits.length - 1] + 1;
    
    
    var startEndId = this.vorbits.findIndex((element) => element === this.vorbits[vertexIds[0]]);
    var difference = vertexIds[0] - startEndId;
    var shiftDifferences = this.getShiftDifference(vertexIds);
    
    var faceVertexIds = [];
    for (let i = 0; i < vertexIds.length; i++)
      faceVertexIds.push(vertexIds[i] - difference);
    
    for (let i = 0; i < rangeX * rangeY; i++) {
      //if vertex orbits do not match
      if (Math.max(...faceVertexIds) >= this.vertices.length)
        break;
      if (Math.min(...faceVertexIds) >= 0 && this.equalShiftDifferences(shiftDifferences, faceVertexIds) && this.equalVOrbits(vertexIds, faceVertexIds)) {
        this.faces.push(faceVertexIds);
        this.forbits.push(orbit);
      }
      //Increase all face vertex Ids by 1
      faceVertexIds = faceVertexIds.map(v=> v+1);
    }
    //console.log("faces: " + this.faces);
  }
  
  /*==========================================================================================*/
  /*================================= HELPERS FOR V DELETION =================================*/
  /*==========================================================================================*/
  removeVOrbitFromEdgeData(orbit, fSize) {
    for (let j = this.edges.length - 1; j > -1; j--) {
      var edge = this.edges[j];
      if (this.vorbits[edge[0]] === orbit || this.vorbits[edge[1]] === orbit) {
        this.edges.splice(j, 1);
        this.eorbits.splice(j,1);
      }
    }
    this.updateEdgeVertices(orbit, fSize);
    this.reassignEdgeOrbits();
  }
  updateEdgeVertices(removedVOrbit, fSize) {
    var numCopies = (fSize.maxx - fSize.minx + 1) * (fSize.maxy - fSize.miny + 1);
    for (let j = this.edges.length - 1; j > -1; j--) {
      var edge = this.edges[j];
      if (this.vorbits[edge[0]] > removedVOrbit) {
        edge[0] = edge[0] - numCopies;
      }
      if (this.vorbits[edge[1]] > removedVOrbit) {
        edge[1] = edge[1] - numCopies;
      }
    }
  }
  reassignEdgeOrbits() {
    var currOrbitValue = this.eorbits[0];
    var correctOrbitValue = 0;//the first value of eorbits should be 0.
    for (let j = 0; j < this.eorbits.length; j++) {
      if (this.eorbits[j] != currOrbitValue) {
        currOrbitValue = this.eorbits[j];
        correctOrbitValue++;
      }
      this.eorbits[j] = correctOrbitValue;
    } 
  }
  removeVOrbitFromEdgeRepData(orbit) {
    var ogLength = this.edgeReps.length - 1;
    for (let j = this.edgeReps.length - 1; j > -1; j--) {
      var edgeRep = this.edgeReps[j];
      if (edgeRep[0] === orbit || edgeRep[1] === orbit) {
        this.edgeReps.splice(j, 1);
        this.eRepShifts.splice(j, 1);
        var orbitColor = EDGE_COLORS[j];
        for (let i = j; i <= ogLength; i++) {
          EDGE_COLORS[i] = EDGE_COLORS[i+1];
        }
        EDGE_COLORS[ogLength] = orbitColor;
        ogLength--;
      }
    }
    this.updateEdgeRepVertices(orbit);
  }
  updateEdgeRepVertices(orbit) {
    for (let j = this.edgeReps.length - 1; j > -1; j--) {
      var edgeRep = this.edgeReps[j];
      if (edgeRep[0] > orbit) {
        edgeRep[0] = edgeRep[0] - 1;
      }
      if (edgeRep[1] > orbit) {
        edgeRep[1] = edgeRep[1] - 1;
      }
    }
  }
  removeVOrbitFromVertexData(orbit) {
    for (let i = this.vertices.length - 1; i > -1; i--) {
      if (this.vorbits[i] === orbit) {
        this.vertices.splice(i, 1); 
        this.vorbits.splice(i, 1);
        this.vshifts.splice(i,1);
      }
    }
    this.vertexRepsFrac.splice(orbit,1);
    this.reassignVertexOrbits(orbit);
  }
  reassignVertexOrbits(orbit) {
    let orbitColor = VERTEX_COLORS[orbit];
    for (let i = this.vertices.length - 1; i > -1; i--) {
      if (this.vorbits[i] > orbit) {
        this.vorbits[i] = this.vorbits[i] - 1;
      }
    }
    for (let i = orbit; i <= this.vertexRepsFrac.length; i++) {
      VERTEX_COLORS[i] = VERTEX_COLORS[i+1];
    }
    VERTEX_COLORS[this.vertexRepsFrac.length] = orbitColor;
  }
  /*===============================================================================*/
  /*================================= FOR RESIZING =================================*/
  /*================================================================================*/
  makeNewFragment(fSize, updateEdges = true) {
    this.vertices = [];
    this.vorbits = [];
    this.vshifts = [];
    var cartCoords = fracToCartCoords(this, this.vertexRepsFrac);
    for (let i = 0; i < cartCoords.length; i++) {
      var repCoord = getVertexRepCoord(this, cartCoords[i]);
      this.computeVertexCopies(repCoord, i, fSize);
    }
    
    if (updateEdges) {
      this.edges = [];
      this.eorbits = [];
      for (let i = 0; i < this.edgeReps.length; i++) {
        this.computeEdgeCopies(this.edgeReps[i][0], this.edgeReps[i][1], this.eRepShifts[i], i, fSize);
      }
    }
  }
  /*===========================================================================*/
  /*================================ RECOMPUTE ================================*/
  /*===========================================================================*/
  recomputeVertexReps() {
    this.vertexRepsFrac = [];
    var visitedOrbit = [];
    for (let i = 0; i <= Math.max(...this.vorbits); i++) {
      visitedOrbit.push(false);
    }
    var k = 0;
    while (visitedOrbit.includes(false)) {
      let orbit = this.vorbits[k];
      if (!visitedOrbit[orbit]) {
        this.vertexRepsFrac.push(cartToFracCoord(this, this.vertices[k], true));
        visitedOrbit[orbit] = true;
        if (this.vertexRepsFrac.length > VERTEX_COLORS.length) {
          //https://stackoverflow.com/questions/1152024/best-way-to-generate-a-random-color-in-javascript
          var randomColor = '#' + (0x1000000 + Math.floor(Math.random() * 0x1000000)).toString(16).substr(1);
          VERTEX_COLORS.push(randomColor);
        }
      }
      k++;
    }
  } 
  recomputeRelaxedVertexReps() {
    this.vertexRepsFrac = [];
    for (let i = 0; i < this.vertices.length; i++) {
      var fracPt = cartToFracCoord(this, this.vertices[i]);
      if (fracPt.x >= 0 && fracPt.x < 1 && fracPt.y >= 0 && fracPt.y < 1) {
        this.vertexRepsFrac.push(fracPt);
        if (this.vertexRepsFrac.length > VERTEX_COLORS.length) {
          //https://stackoverflow.com/questions/1152024/best-way-to-generate-a-random-color-in-javascript
          var randomColor = '#' + (0x1000000 + Math.floor(Math.random() * 0x1000000)).toString(16).substr(1);
          VERTEX_COLORS.push(randomColor);
        }
      }
    }
  } 
  recomputeEdgeRepShifts() {
    for (let i = 0 ; i < this.edgeReps.length; i++) {
      var index = this.eorbits.findIndex((element) => element === i);
      var edge = this.edges[index];
      var v1Shift = getVertexShift(this, this.vertices[edge[0]]);
      var v2Shift = getVertexShift(this, this.vertices[edge[1]]);
      v2Shift.subtract(v1Shift);
      this.eRepShifts[i] = v2Shift;
    }
  }
  //Compute edgeReps and eRepShifts
  recomputeEdgeReps() {
    var visitedOrbit;
    (visitedOrbit = []).length = Math.max(...this.eorbits) + 1;
    visitedOrbit.fill(false);
    var k = 0;
    while (visitedOrbit.includes(false)) {
      let orbit = this.eorbits[k];
      if (!visitedOrbit[orbit]) {
        var edgeRep = [this.vorbits[this.edges[k][0]], this.vorbits[this.edges[k][1]]];
        this.edgeReps.push(edgeRep);
        if (EDGE_COLORS.length < this.edgeReps.length)
          EDGE_COLORS.push('#' + (0x1000000 + Math.floor(Math.random() * 0x1000000)).toString(16).substr(1));
        var shift = new Coordinate(this.vshifts[this.edges[k][1]].x,this.vshifts[this.edges[k][1]].y);
        shift.subtract(this.vshifts[this.edges[k][0]]);
        this.eRepShifts.push(shift);
        visitedOrbit[orbit] = true;
      }
      k++;
    }
  }
  recomputeRelaxedEdgeReps() {
    var oldNrReps = this.edgeReps.length;
    this.edgeReps = [];
    this.eRepShifts = [];
    var edgeInfo = [];
    for (let i = 0; i < this.edges.length; i++) {
      var v1 = this.vertices[this.edges[i][0]];
      var v2 = this.vertices[this.edges[i][1]];
      var edgeInfo = getEdgeRep(this, v1, v2);
      if (!this.inEdgeReps(edgeInfo[0], edgeInfo[1])) {
        this.edgeReps.push(edgeInfo[0]);
        this.eRepShifts.push(edgeInfo[1]);
        if (oldNrReps < this.edgeReps.length)
          EDGE_COLORS.push('#' + (0x1000000 + Math.floor(Math.random() * 0x1000000)).toString(16).substr(1));
      }
    }
  }
  inEdgeReps(edge, shift) {
    for (let i = 0; i < this.edgeReps.length; i++) {
      var eRep = this.edgeReps[i];
      var eRepShift = this.eRepShifts[i];
      if (equalEdges(edge, shift, eRep, eRepShift))
        return true;
    }
    return false;
  }
  /*===============================================================================*/
  /*================================ PERTURBATIONS ================================*/
  /*===============================================================================*/
  perturbAxes(){
    var random = getRandom([1,2]);
    
    for (let i = 0; i < random; i++) {
      var axis = this.axes[i];//axis that will be perturbed
      var randomx = Math.random() < 0.5;//decide if the x value of the axis will be perturbed
      var randomy = Math.random() < 0.5;//decide if the y value of the axis will be perturbed
      while (!randomx && !randomy) {//at least one must be true
        randomx = Math.random() < 0.5;
        randomy = Math.random() < 0.5;
      }
      if (randomx)
        axis.x += getRandom(PERTURB_RANGE);//perturb
      if (randomy)
        axis.y += getRandom(PERTURB_RANGE);//perturb
    }
  }
  perturbVOrbits(){
    var orbits = getMultipleRandom([...Array(this.vertexRepsFrac.length).keys()], getRandomInt([1,this.vertexRepsFrac.length]));
    
    for (let i = 0; i < orbits.length; i++) {
      var vertexRep = this.vertexRepsFrac[orbits[i]];
      var randomx = Math.random() < 0.5;//decide if the x value of the axis will be perturbed
      var randomy = Math.random() < 0.5;//decide if the y value of the axis will be perturbed
      while (!randomx && !randomy) {//at least one must be true
        randomx = Math.random() < 0.5;
        randomy = Math.random() < 0.5;
      }
      if (randomx)
        this.vertexRepsFrac[orbits[i]].x += getRandom(PERTURB_FRAC_RANGE);//perturb
      if (randomy)
        this.vertexRepsFrac[orbits[i]].y += getRandom(PERTURB_FRAC_RANGE);//perturb
    }
    
  }
}